//
//  SvmTrainer.cpp
//  SvmTrainer
//
//  Created by 熊田 聖也 on 12/05/26.
//  Copyright (c) 2012年 __MyCompanyName__. All rights reserved.
//

#include "SvmTrainer.h"
#include <shogun/features/Labels.h>
#include <shogun/features/SimpleFeatures.h>
#include <shogun/kernel/GaussianKernel.h>
#include <shogun/kernel/Chi2Kernel.h>
#include <shogun/kernel/HistogramIntersectionKernel.h>
#include <shogun/classifier/svm/LibSVM.h>
#include <shogun/base/init.h>
#include <shogun/lib/common.h>
#include <shogun/io/SGIO.h>
#include <shogun/io/SerializableAsciiFile.h>
#include <opencv2/imgproc/imgproc.hpp>
#include "cvmat_serialization.h"
#include <cstdio>
#include <map>

namespace 
{
        void print_message(FILE* target, const char* str)
        {
                fprintf(target, "%s", str);
        }
        
        void print_warning(FILE* target, const char* str)
        {
                print_message(target, str);
        }
        
        void print_error(FILE* target, const char* str)
        {
                print_message(target, str);
        }

        shogun::CDotKernel* get_gaussian_kernel(const StParameters& params)
        {
                return new shogun::CGaussianKernel(params.kernel_cache_size_, params.width_);
        }
        
        shogun::CDotKernel* get_chi2_kernel(const StParameters& params)
        {
                return new shogun::CChi2Kernel(params.kernel_cache_size_, params.width_);
        }
        
        shogun::CDotKernel* get_histogram_intersection_kernel(const StParameters& params)
        {
                shogun::CHistogramIntersectionKernel* kernel = new shogun::CHistogramIntersectionKernel(params.kernel_cache_size_);
                kernel->set_beta(params.beta_);
                return kernel;
        }
        
}

SvmTrainer::SvmTrainer(const StParameters& params)
        : params_(params),
          svm_(0)
{
        get_kernel_["chi2"] = get_chi2_kernel;
        get_kernel_["histogram-intersection"] = get_histogram_intersection_kernel;
        get_kernel_["gaussian"] = get_gaussian_kernel;
}
        
SvmTrainer::~SvmTrainer()
{
        if ( svm_ != 0 ) {
                SG_UNREF(svm_);
        }
}

namespace
{
        // test ok
        shogun::CSimpleFeatures<float64_t>* create_features_with_float64_t(const cv::Mat& positive, const cv::Mat& negative)
        {
                assert(positive.type() == CV_32FC1);
                const int prows = positive.rows;
                const int nrows = negative.rows;
                const int cols = positive.cols;
                const int rows = prows + nrows;
                const int size = rows * cols;
                
                const float32_t* pptr = reinterpret_cast<const float32_t*>(positive.data);
                const float32_t* nptr = reinterpret_cast<const float32_t*>(negative.data);
                
                float64_t* matrix = SG_MALLOC(float64_t, size);
                std::copy(pptr, pptr + prows * cols, matrix);
                std::copy(nptr, nptr + nrows * cols, matrix + prows * cols);
                
                // create three cols-dimensional vectors
                // shogun will now own the matrix created
                shogun::CSimpleFeatures<float64_t>* features = new shogun::CSimpleFeatures<float64_t>();
                features->set_feature_matrix(matrix, cols, rows); // num_features(dimension), num_vectors
                return features;
        }
        
        // test ok
        template<typename T>
        void set_labels(
                const shogun::CSimpleFeatures<T>*       features, 
                int                                     positive_rows, 
                shogun::CLabels*                        labels
        )
        {
                for ( int i = 0; i < positive_rows; ++i ) {
                        labels->set_label(i, 1);
                }
                
                for ( int i = positive_rows, n = features->get_num_vectors(); i < n; ++i ) {
                        labels->set_label(i, -1);
                }
        }
}


// test ok
void SvmTrainer::execute(const cv::Mat& positive, const cv::Mat& negative)
{
        assert(positive.type() == cv::DataType<float32_t>::type);
        shogun::CSimpleFeatures<float64_t>* features = create_features_with_float64_t(positive, negative);
        assert(features->get_num_features() == positive.cols);
        assert(features->get_num_vectors() == positive.rows + negative.rows);
        
        // create labels
        shogun::CLabels* labels = new shogun::CLabels(features->get_num_vectors());
        set_labels<float64_t>(features, positive.rows, labels);
        
        shogun::CDotKernel* kernel = get_kernel_[params_.kernel_type_](params_);
        kernel->init(features, features);
        
        // create libsvm and train
        svm_ = new shogun::CLibSVM(params_.cost_, kernel, labels);
        svm_->train();
}

bool SvmTrainer::load(const std::string& svm_filename)
{       
        if ( svm_ != 0 ) {
                SG_UNREF(svm_);
        }
                
        svm_ = new shogun::CLibSVM();
        shogun::CSerializableAsciiFile svm_file(const_cast<char*>(svm_filename.c_str()), 'r');
        if ( !svm_->load_serializable(&svm_file) ) {
                std::cout << "unable to load svm file\n";
                return false;
        }
        return true;
}

bool SvmTrainer::save(const std::string& svm_filename) const
{
        shogun::CSerializableAsciiFile svm_file(const_cast<char*>(svm_filename.c_str()), 'w');
        if ( !svm_->save_serializable(&svm_file) ) {
                std::cout << "unable to save to svm file\n";
                return false;
        }
        return true;
}

shogun::CLibSVM* SvmTrainer::get_svm()
{
        return svm_;
}


#ifdef UNIT_TEST
#include <boost/test/unit_test.hpp>
#include <boost/random.hpp>
#include <cstdio>
#include <shogun/lib/Time.h>
#include "common.h"
#include <boost/test/floating_point_comparison.hpp>

namespace
{
        template<typename T>
        shogun::CSimpleFeatures<T>* create_features(int rows, int cols)
        {
                int size = rows * cols;
                // create some data
                T* matrix = SG_MALLOC(T, size);
                for ( int32_t i = 0; i < size; ++i ) {
                        matrix[i] = i;
                }

                
                // create three cols-dimensional vectors
                // shogun will now own the matrix created
                shogun::CSimpleFeatures<T>* features = new shogun::CSimpleFeatures<T>();
                features->set_feature_matrix(matrix, cols, rows); 
                return features;
        }

        template<typename T>
        void test_set_labels()
        {
                StParameters params;
                SvmTrainer trainer(params);
                int rows = 9;
                int cols = 5;
                cv::Mat m(rows, cols, CV_32FC1);
                const shogun::CSimpleFeatures<T>* features = create_features<T>(rows, cols);
                shogun::CLabels* labels = new shogun::CLabels(features->get_num_vectors());
                set_labels<T>(features, 3, labels);
                
                BOOST_CHECK(2 == labels->get_num_classes());
                BOOST_CHECK(9 == labels->get_num_labels());
                
                for ( int i = 0; i < 3; ++i ) {
                        BOOST_CHECK(1 == labels->get_label(i));
                }
                
                for ( int i = 3; i < 9; ++i ) {
                        BOOST_CHECK(-1 == labels->get_label(i));
                }
        }
        
        template<typename T>
        void set_positive_values(cv::Mat& m)
        {
                assert(cv::DataType<T>::type == m.type());
                boost::mt19937 gen;
                boost::uniform_real<> dst(10, 100);
                boost::variate_generator<boost::mt19937, boost::uniform_real<> > rnd(gen, dst);
                cv::MatIterator_<T> first = m.begin<T>();
                cv::MatIterator_<T> last = m.end<T>();
                while ( first != last ) {
                        *first = rnd();
                        ++first;
                }
        }
        
        template<typename T>
        void set_negative_values(cv::Mat& m)
        {
                assert(cv::DataType<T>::type == m.type());
                boost::mt19937 gen;
                boost::uniform_real<> dst(-100, -10);
                boost::variate_generator<boost::mt19937, boost::uniform_real<> > rnd(gen, dst);
                cv::MatIterator_<T> first = m.begin<T>();
                cv::MatIterator_<T> last = m.end<T>();
                while ( first != last ) {
                        *first = rnd();
                        ++first;
                }
        
        }
        
        
template<typename T>
        void set_test_positive_values(cv::Mat& m)
        {
                assert(cv::DataType<T>::type == m.type());
                boost::mt19937 gen;
                boost::uniform_real<> dst(20, 50);
                boost::variate_generator<boost::mt19937, boost::uniform_real<> > rnd(gen, dst);
                cv::MatIterator_<T> first = m.begin<T>();
                cv::MatIterator_<T> last = m.end<T>();
                while ( first != last ) {
                        *first = rnd();
                        ++first;
                }
        }
        
        template<typename T>
        void apply_on_test(SvmTrainer& trainer, int rows, int cols)
        {
                cv::Mat test(rows / 2, cols, cv::DataType<T>::type);
                set_test_positive_values<T>(test);
                shogun::CSimpleFeatures<float64_t>* test_features = new shogun::CSimpleFeatures<float64_t>();
                common::convert(test, test_features);
                
                shogun::CLabels* labels = trainer.get_svm()->apply(test_features);
                BOOST_CHECK(rows / 2 == labels->get_num_labels());
                
                for ( int i = 0; i < rows / 2; ++i ) {
                        BOOST_CHECK(0 < labels->get_label(i));
                }
                
                SG_UNREF(labels);
        }
     
        template<typename T>
        void test_execute(const std::string& kernel)
        {
                using namespace shogun;
//                CTime t;
//                t.io->set_loglevel(MSG_INFO);

                int rows = 200;
                int cols = 2;
                cv::Mat positive(rows/2, cols, cv::DataType<T>::type);
                cv::Mat negative(rows/2, cols, cv::DataType<T>::type);
                set_positive_values<T>(positive);
                set_negative_values<T>(negative);
                
                StParameters params;
                params.kernel_type_ = kernel;
                params.cost_ = 10;
                SvmTrainer trainer(params);
                trainer.execute(positive, negative);
                
                // classify on training examples
                for ( int32_t i = 0; i < rows; ++i ) {
                        double index = trainer.get_svm()->apply(i);
                        if ( i < rows / 2 ) {
                                BOOST_CHECK(index > 0.0);
                        } else {
                                BOOST_CHECK(index < 0.0);
                        }
                }
//                std::cout << "test_execute\n";
                apply_on_test<T>(trainer, rows, cols);
        }
        

        
        template<typename T>
        void test_save(const std::string& kernel)
        {
                const std::string svm_filename = "/Users/kumada/Projects/scenerecognizer/SvmTrainer/SvmTrainer/src/svm.bin";
                const std::string features_filename = "/Users/kumada/Projects/scenerecognizer/SvmTrainer/SvmTrainer/src/features.bin";
        
                using namespace shogun;
//                CTime t;
//                t.io->set_loglevel(MSG_INFO);
//                
                int rows = 200;
                int cols = 2;
                cv::Mat positive(rows/2, cols, cv::DataType<T>::type);
                cv::Mat negative(rows/2, cols, cv::DataType<T>::type);
                set_positive_values<T>(positive);
                set_negative_values<T>(negative);
                
                StParameters params;
                params.kernel_type_ = kernel;
                params.cost_ = 10;
                SvmTrainer trainer(params);
                trainer.execute(positive, negative);
                BOOST_CHECK(trainer.save(svm_filename));
        }
        
        
        
        void test_load() 
        {
                const std::string svm_filename = "/Users/kumada/Projects/scenerecognizer/SvmTrainer/SvmTrainer/src/svm.bin";
                const std::string features_filename = "/Users/kumada/Projects/scenerecognizer/SvmTrainer/SvmTrainer/src/features.bin";
//                shogun::CTime t;
//                t.io->set_loglevel(shogun::MSG_INFO);

                SvmTrainer trainer; // default constructorを作るべき。
                BOOST_CHECK(trainer.load(svm_filename));//, features_filename));
                
                int rows = trainer.get_svm()->get_kernel()->get_num_vec_lhs();
                int r = trainer.get_svm()->get_kernel()->get_num_vec_rhs();
                BOOST_CHECK(200 == rows);
                BOOST_CHECK(200 == r);
                
                
                // classify on training examples
                for ( int32_t i = 0; i < rows; ++i ) {
                        double index = trainer.get_svm()->apply(i);
                        if ( i < rows / 2 ) {
                                BOOST_CHECK(index > 0.0);
                        } else {
                                BOOST_CHECK(index < 0.0);
                        }
                }
                
                int cols = 2;
                std::cout << "test_load\n";
                apply_on_test<float32_t>(trainer, rows, cols);
        }
        
        void check_sum(const cv::Mat& m)
        {
                const double tolerance = 1.0e-05;
                int cols = m.cols;
                int rows = m.rows;
                for ( int j = 0; j < rows; ++j ) {
                        const float* p = m.ptr<float>(j);
                        BOOST_CHECK_CLOSE(1.0, std::accumulate(p, p + cols, 0.0), tolerance);
//                        double d = 1.0f - std::accumulate(p, p + cols, 0.0);
//                        std::cout << d << std::endl;
                        
                }
        }

        void test_execute_with_real_data(const std::string& kernel)
        {
                std::cout << "kernel: " << kernel << std::endl;
                using namespace shogun;
                StParameters params;
                params.input_positive_ = "/Users/kumada/Data/scene-recognizer/histograms/CALsuburb_histograms.bin";
                params.input_negative_ = "/Users/kumada/Data/scene-recognizer/histograms/MITcoast_histograms.bin";

                cv::Mat positive;
                load_mat(positive, params.input_positive_);
                cv::Mat negative;
                load_mat(negative, params.input_negative_);
                check_sum(positive);
                params.kernel_type_ = kernel;
                params.cost_ = 10;
                SvmTrainer trainer(params);
                trainer.execute(positive, negative);
                
                int prows = positive.rows;
                int nrows = negative.rows;
                BOOST_CHECK(241 == prows);
                BOOST_CHECK(360 == nrows);
                BOOST_CHECK(601 == prows + nrows);
                // classify on training examples
                for ( int32_t i = 0; i < prows + nrows; ++i ) {
                        double index = trainer.get_svm()->apply(i);
                        if ( i < prows ) {
                                BOOST_CHECK(index > 0.0);
                        } else {
                                BOOST_CHECK(index < 0.0);
                        }
                }
        }

        
        void test_save_with_real_data(const std::string& kernel)
        {
                using namespace shogun;
//                CTime t;
//                t.io->set_loglevel(MSG_INFO);
//   
                StParameters params;
                params.input_positive_ = "/Users/kumada/Data/scene-recognizer/histograms/CALsuburb_histograms.bin";
                params.input_negative_ = "/Users/kumada/Data/scene-recognizer/histograms/MITcoast_histograms.bin";

                cv::Mat positive;
                load_mat(positive, params.input_positive_);
                cv::Mat negative;
                load_mat(negative, params.input_negative_);

                params.kernel_type_ = kernel;
                params.cost_ = 10;
                SvmTrainer trainer(params);
                trainer.execute(positive, negative);
                params.output_svm_ = "/Users/kumada/Projects/scenerecognizer/SvmTrainer/SvmTrainer/src/real_svm.bin";
                trainer.save(params.output_svm_);
                
        }
        
        void test_load_with_real_data()
        {
                const std::string svm_filename = "/Users/kumada/Projects/scenerecognizer/SvmTrainer/SvmTrainer/src/real_svm.bin";
                const std::string features_filename = "/Users/kumada/Projects/scenerecognizer/SvmTrainer/SvmTrainer/src/real_features.bin";

                SvmTrainer trainer;
                BOOST_CHECK(trainer.load(svm_filename));//, features_filename));
                
                int prows = 241;
                int nrows = 360;
                int rows = prows + nrows;
                
                // classify on training examples
                for ( int32_t i = 0; i < rows; ++i ) {
                        double index = trainer.get_svm()->apply(i);
                        if ( i < prows ) {
                                BOOST_CHECK(index > 0.0);
                        } else {
                                BOOST_CHECK(index < 0.0);
                        }
                }
        }
        
        void test_create_features_with_float64_t()
        {
                StParameters params;
                SvmTrainer trainer(params);
                int cols = 2;
                int rows = 100;
                int nrows = 101;
                cv::Mat positive(rows, cols, cv::DataType<float32_t>::type);
                cv::Mat negative(nrows, cols, cv::DataType<float32_t>::type);
                set_positive_values<float32_t>(positive);
                set_negative_values<float32_t>(negative);
                
                for ( int j = 0; j < positive.rows; ++j ) {
                        const float32_t* p = positive.ptr<float32_t>(j);
                        for ( int i = 0; i < positive.cols; ++i ) {
                                BOOST_CHECK(p[i] > 0.0f);
                        }
                }
                for ( int j = 0; j < negative.rows; ++j ) {
                        const float32_t* p = negative.ptr<float32_t>(j);
                        for ( int i = 0; i < negative.cols; ++i ) {
                                BOOST_CHECK(p[i] < 0.0f);
                        }
                }
                
                shogun::CSimpleFeatures<float64_t>* features = create_features_with_float64_t(positive, negative);
                for ( int j = 0, h = positive.rows; j < h; ++j ) {
                        const float32_t* p = positive.ptr<float32_t>(j);
                        const shogun::SGVector<float64_t>& q = features->get_feature_vector(j);
                        for ( int i = 0, w = positive.cols; i < w; ++i ) {
                                BOOST_CHECK(p[i] == q[i]);
                        }
                }
                
                for ( int j = 0, h = negative.rows; j < h; ++j ) {
                        const float32_t* p = negative.ptr<float32_t>(j);
                        const shogun::SGVector<float64_t>& q = features->get_feature_vector(j + rows);
                        for ( int i = 0, w = negative.cols; i < w; ++i ) {
                                BOOST_CHECK(p[i] == q[i]);
                        }
                }

                SG_UNREF(features);
        
        }
        
}

BOOST_AUTO_TEST_CASE(TEST_SvmTrainer)
{
        std::cout << "TEST_SvmTrainer:begin\n";
        
        shogun::init_shogun(&print_message, &print_warning, &print_error);
        test_create_features_with_float64_t();
        test_set_labels<float>();
        test_set_labels<double>();
        test_save<float>("histogram-intersection");
        test_load();
        test_execute<float>("histogram-intersection");
        test_execute_with_real_data("histogram-intersection");
        test_execute_with_real_data("chi2");
        test_save_with_real_data("histogram-intersection");
        test_load_with_real_data();
        test_save_with_real_data("chi2");
        test_load_with_real_data();
        
        shogun::exit_shogun();
        std::cout << "TEST_SvmTrainer:end\n";
}

#endif // UNIT_TEST



