//
//  main.cpp
//  SvmTrainer
//
//  Created by 熊田 聖也 on 12/05/21.
//  Copyright (c) 2012年 __MyCompanyName__. All rights reserved.
//

#include <iostream>
#include <boost/program_options.hpp>
#include "StParameters.h"
#include "common.h"
#include "SvmTrainer.h"
#include "cvmat_serialization.h"
#include <shogun/base/init.h>
#include <boost/filesystem.hpp>

namespace fs = boost::filesystem;
namespace po = boost::program_options;

namespace
{
        void set_params(po::variables_map& vm, StParameters& param); 
}

#ifndef UNIT_TEST
int main(int argc, const char * argv[])
{
        shogun::init_shogun_with_defaults();
        try {
                StParameters params;
                po::options_description desc("options"); 
                desc.add_options()
                        ("help", "show this help")
                        ("input-positive", po::value<std::string>(), "directory path to positive training data")
                        ("input-negative", po::value<std::string>(), "directory path to negative training data")
                        ("output-svm", po::value<std::string>()->default_value(params.output_svm_), "path to output file for svm")
                        ("kernel-type", po::value<std::string>()->default_value(params.kernel_type_), "gaussian, chi2 or histogram-intersection")
                        ("kernel-cache-size", po::value<int>()->default_value(params.kernel_cache_size_), "kernel cache size(MB)")
                        ("width", po::value<double>()->default_value(params.width_), "width used for gaussian and ch2kernel")
                        ("cost", po::value<double>()->default_value(params.cost_), "parameter C of C-SVC")
                        ("beta", po::value<double>()->default_value(params.beta_), "parameter beta of histogram-intersection")
                ;

                po::variables_map vm;
                po::store(po::parse_command_line(argc, argv, desc), vm);
                po::notify(vm);

                if ( !vm.count("input-positive") || !vm.count("input-negative") || vm.count("help") ) {
                        std::cout << desc << std::endl;
                       return 1;
                }
                set_params(vm, params);
                std::cout << params << std::endl;
                
                cv::Mat positive;
                common::accumulate_cvmats(params.input_positive_, positive);
                
                cv::Mat negative;
                common::accumulate_cvmats(params.input_negative_, negative);
                
                SvmTrainer trainer(params);
                trainer.execute(positive, negative);
                trainer.save(params.output_svm_);
                
        } catch (const std::exception& e) {
                std::cout << e.what() << std::endl;
        }
        shogun::exit_shogun();
        return 0;
}
#endif // UNIT_TEST

namespace
{
        void set_params(po::variables_map& vm, StParameters& params)
        {
                common::set_params(vm, "input-positive", params.input_positive_);
                common::set_params(vm, "input-negative", params.input_negative_);
                common::set_params(vm, "output-svm", params.output_svm_);
                common::set_params(vm, "kernel-type", params.kernel_type_);
                common::set_params(vm, "kernel-cache-size", params.kernel_cache_size_);
                common::set_params(vm, "width", params.width_);
                common::set_params(vm, "cost", params.cost_);
                common::set_params(vm, "beta", params.beta_);
        }
}

#ifdef UNIT_TEST
#define BOOST_AUTO_TEST_MAIN
#include <boost/test/unit_test.hpp>
#include <shogun/kernel/GaussianKernel.h>
#include <shogun/kernel/Chi2Kernel.h>
#include <shogun/classifier/svm/LibSVM.h>
#include <shogun/io/SerializableAsciiFile.h>
#include <shogun/lib/Time.h>
#include <boost/lexical_cast.hpp>

namespace
{
        void print_message(FILE* target, const char* str)
        {
                fprintf(target, "%s", str);
        }
        
        void print_warning(FILE* target, const char* str)
        {
                print_message(target, str);
        }
        
        void print_error(FILE* target, const char* str)
        {
                print_message(target, str);
        }
        
        using namespace shogun;

        template<typename T>
        shogun::CSimpleFeatures<T>* create_features()
        {
                // create some data
                T* matrix = SG_MALLOC(T, 6);
                for ( int32_t i = 0; i < 6; ++i ) {
                        matrix[i] = i;
                }
                
                // create three 2-dimensional vectors
                // shogun will now own the matrix created
                CSimpleFeatures<T>* features = new CSimpleFeatures<T>();
                features->set_feature_matrix(matrix, 2, 3); // 2 is cols
                return features;
        }
        
        template<typename T>
        void test_shogun()
        {
                CSimpleFeatures<T>* features = create_features<T>();
                
                // create three labels
                CLabels* labels = new CLabels(3);
                labels->set_label(0, -1);
                labels->set_label(1, +1);
                labels->set_label(2, -1);
                
                // create gaussian kernel with cache 10MB, width = 0.5
                //CGaussianKernel* kernel = new CGaussianKernel(10, 0.5);
                CDotKernel* kernel = new CChi2Kernel(100, 0.5);
                //CHistogramIntersectionKernel* kernel = new CHistogramIntersectionKernel(10);
                
                kernel->init(features, features);
                
                // create libsvm with C = 10 and train
                CLibSVM* svm = new CLibSVM(10, kernel, labels);
                svm->train();
                //classify on training examples
                BOOST_CHECK(svm->apply(0) < 0.0);
                BOOST_CHECK(svm->apply(1) > 0.0);
                BOOST_CHECK(svm->apply(2) < 0.0);
        }
        
        const char* svm_filename = "/Users/kumada/Projects/scenerecognizer/SvmTrainer/SvmTrainer/src/svm_main.bin";
        const char* features_filename = "/Users/kumada/Projects/scenerecognizer/SvmTrainer/SvmTrainer/src/features_main.bin";
        
        template<typename T>
        void test_shogun_save()
        {
                CSimpleFeatures<T>* features = create_features<T>();
                
                // create three labels
                CLabels* labels = new CLabels(3);
                labels->set_label(0, -1);
                labels->set_label(1, +1);
                labels->set_label(2, -1);
                
                CDotKernel* kernel = new CChi2Kernel(100, 0.5);
                kernel->init(features, features);
                
                // create libsvm with C = 10 and train
                CLibSVM* svm = new CLibSVM(10, kernel, labels);
                svm->train();
                
                CSerializableAsciiFile* svm_file = new CSerializableAsciiFile(const_cast<char*>(svm_filename), 'w');
                svm->save_serializable(svm_file);
                SG_UNREF(svm_file);
                
                CSerializableAsciiFile* features_file = new CSerializableAsciiFile(const_cast<char*>(features_filename), 'w');
                features->save_serializable(features_file);
                SG_UNREF(features_file);
                
                                
                // free up memory: why does the only svm explicitly call SG_UNREF?
                SG_UNREF(svm);
        }

        void test_shogun_load()
        {
                CTime t;
                t.io->set_loglevel(MSG_ERROR);
                
                CLibSVM* svm = new CLibSVM();
                CSerializableAsciiFile svm_file(const_cast<char*>(svm_filename), 'r');
                if ( !svm->load_serializable(&svm_file) ) {
                        std::cout << "unable to load file\n";
                        return;
                }
                
                CSimpleFeatures<float64_t>* features = new CSimpleFeatures<float64_t>();
                CSerializableAsciiFile features_file(const_cast<char*>(features_filename), 'r');
                if ( !features->load_serializable(&features_file) ) {
                        std::cout << "unable to load file\n";
                        return;
                }
                
                svm->get_kernel()->init(features, features);
                CLabels* out_labels = svm->apply(features);
                
                BOOST_CHECK(3 == out_labels->get_num_labels());
                BOOST_CHECK(out_labels->get_label(0) < 0);
                BOOST_CHECK(out_labels->get_label(1) > 0);
                BOOST_CHECK(out_labels->get_label(2) < 0);
                
                SG_UNREF(svm);
                SG_UNREF(features);
        }
        
        void set_values(cv::Mat& m, int i)
        {
                cv::MatIterator_<double> first = m.begin<double>();
                cv::MatIterator_<double> last = m.end<double>();
                while ( first != last ) {
                        *first = i;
                        ++first;
                }
        }
}

BOOST_AUTO_TEST_CASE(TEST_main)
{       
        std::cout << "TEST_main:begin\n";
        init_shogun(&print_message, &print_warning, &print_error);
        test_shogun_save<float64_t>();
        test_shogun_load();
        exit_shogun();
        std::cout << "TEST_main:end\n";
}
#endif // UNIT_TEST
